@@ -3,8 +3,9 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2013-2019, VU University Amsterdam
+    Copyright (c)  2013-2023, VU University Amsterdam
                               CWI, Amsterdam
+                              SWI-Prolog Solutions b.v
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -37,12 +38,18 @@
           [ safe_goal/1,                % :Goal
             safe_call/1                 % :Goal
           ]).
-:- use_module(library(assoc)).
-:- use_module(library(lists)).
-:- use_module(library(debug)).
-:- use_module(library(error)).
-:- use_module(library(prolog_format)).
-:- use_module(library(apply)).
+:- use_module(library(apply_macros),[expand_phrase/2]).
+:- use_module(library(apply),[maplist/2]).
+:- use_module(library(assoc),[empty_assoc/1,get_assoc/3,put_assoc/4]).
+:- use_module(library(debug),[debug/3,debugging/1]).
+:- use_module(library(error),
+              [ must_be/2,
+                instantiation_error/1,
+                type_error/2,
+                permission_error/3
+              ]).
+:- use_module(library(lists),[append/3]).
+:- use_module(library(prolog_format),[format_types/2]).
 
 :- multifile
     safe_primitive/1,               % Goal
@@ -50,7 +57,8 @@
     safe_meta/2,                    % Goal, Calls
     safe_meta/3,                    % Goal, Context, Calls
     safe_global_variable/1,         % Name
-    safe_directive/1.               % Module:Goal
+    safe_directive/1,               % Module:Goal
+    safe_prolog_flag/2.             % +Name, +Value
 
 % :- debug(sandbox).
 
@@ -420,6 +428,11 @@ term_expansion(safe_primitive(Goal), Term) :-
     ->  Term = safe_primitive(Goal)
     ;   Term = []
     ).
+term_expansion((safe_primitive(Goal) :- Body), Term) :-
+    (   verify_safe_declaration(Goal)
+    ->  Term = (safe_primitive(Goal) :- Body)
+    ;   Term = []
+    ).
 
 system:term_expansion(sandbox:safe_primitive(Goal), Term) :-
     \+ current_prolog_flag(xref, true),
@@ -427,6 +440,12 @@ system:term_expansion(sandbox:safe_primitive(Goal), Term) :-
     ->  Term = sandbox:safe_primitive(Goal)
     ;   Term = []
     ).
+system:term_expansion((sandbox:safe_primitive(Goal) :- Body), Term) :-
+    \+ current_prolog_flag(xref, true),
+    (   verify_safe_declaration(Goal)
+    ->  Term = (sandbox:safe_primitive(Goal) :- Body)
+    ;   Term = []
+    ).
 
 verify_safe_declaration(Var) :-
     var(Var),
@@ -456,13 +475,16 @@ verify_safe_declaration(Goal) :-
     ).
 
 ok_meta(system:assert(_)).
+ok_meta(system:load_files(_,_)).
 ok_meta(system:use_module(_,_)).
 ok_meta(system:use_module(_)).
+ok_meta('$syspreds':predicate_property(_,_)).
 
 verify_predefined_safe_declarations :-
     forall(clause(safe_primitive(Goal), _Body, Ref),
-           ( catch(verify_safe_declaration(Goal), E, true),
-             (   nonvar(E)
+           ( E = error(F,_),
+             catch(verify_safe_declaration(Goal), E, true),
+             (   nonvar(F)
              ->  clause_property(Ref, file(File)),
                  clause_property(Ref, line_count(Line)),
                  print_message(error, bad_safe_declaration(Goal, File, Line))
@@ -497,7 +519,10 @@ safe_primitive(nonvar(_)).
 safe_primitive(system:attvar(_)).
 safe_primitive(integer(_)).
 safe_primitive(float(_)).
+:- if(current_predicate(rational/1)).
 safe_primitive(system:rational(_)).
+safe_primitive(system:rational(_,_,_)).
+:- endif.
 safe_primitive(number(_)).
 safe_primitive(atom(_)).
 safe_primitive(system:blob(_,_)).
@@ -506,6 +531,7 @@ safe_primitive(atomic(_)).
 safe_primitive(compound(_)).
 safe_primitive(callable(_)).
 safe_primitive(ground(_)).
+safe_primitive(system:nonground(_,_)).
 safe_primitive(system:cyclic_term(_)).
 safe_primitive(acyclic_term(_)).
 safe_primitive(system:is_stream(_)).
@@ -540,7 +566,13 @@ safe_primitive(=:=(_,_)).
 safe_primitive(=\=(_,_)).
 safe_primitive(=<(_,_)).
 safe_primitive(<(_,_)).
+:- if(current_prolog_flag(bounded, false)).
 safe_primitive(system:nth_integer_root_and_remainder(_,_,_,_)).
+:- endif.
+safe_primitive(system:current_arithmetic_function(_)).
+safe_primitive(system:bounded_number(_,_,_)).
+safe_primitive(system:float_class(_,_)).
+safe_primitive(system:float_parts(_,_,_,_)).
 
                                         % term-handling
 safe_primitive(arg(_,_,_)).
@@ -548,14 +580,18 @@ safe_primitive(system:setarg(_,_,_)).
 safe_primitive(system:nb_setarg(_,_,_)).
 safe_primitive(system:nb_linkarg(_,_,_)).
 safe_primitive(functor(_,_,_)).
+safe_primitive(system:functor(_,_,_,_)).
 safe_primitive(_ =.. _).
 safe_primitive(system:compound_name_arity(_,_,_)).
 safe_primitive(system:compound_name_arguments(_,_,_)).
 safe_primitive(system:'$filled_array'(_,_,_,_)).
 safe_primitive(copy_term(_,_)).
+safe_primitive(system:copy_term(_,_,_,_)).
 safe_primitive(system:duplicate_term(_,_)).
 safe_primitive(system:copy_term_nat(_,_)).
+safe_primitive(system:size_abstract_term(_,_,_)).
 safe_primitive(numbervars(_,_,_)).
+safe_primitive(system:numbervars(_,_,_,_)).
 safe_primitive(subsumes_term(_,_)).
 safe_primitive(system:term_hash(_,_)).
 safe_primitive(system:term_hash(_,_,_,_)).
@@ -629,6 +665,7 @@ safe_primitive(system:sleep(_)).
 safe_primitive(system:thread_self(_)).
 safe_primitive(system:get_time(_)).
 safe_primitive(system:statistics(_,_)).
+:- if(current_prolog_flag(threads,true)).
 safe_primitive(system:thread_statistics(Id,_,_)) :-
     (   var(Id)
     ->  instantiation_error(Id)
@@ -639,6 +676,7 @@ safe_primitive(system:thread_property(Id,_)) :-
     ->  instantiation_error(Id)
     ;   thread_self(Id)
     ).
+:- endif.
 safe_primitive(system:format_time(_,_,_)).
 safe_primitive(system:format_time(_,_,_,_)).
 safe_primitive(system:date_time_stamp(_,_)).
@@ -655,6 +693,11 @@ safe_primitive(asserta(X)) :- safe_assert(X).
 safe_primitive(assertz(X)) :- safe_assert(X).
 safe_primitive(retract(X)) :- safe_assert(X).
 safe_primitive(retractall(X)) :- safe_assert(X).
+safe_primitive(current_predicate(X)) :- safe_current_predicate(X).
+safe_primitive('$dcg':dcg_translate_rule(_,_)).
+safe_primitive('$syspreds':predicate_property(Pred, _)) :-
+    nonvar(Pred),
+    Pred \= (_:_).
 
 % We need to do data flow analysis to find the tag of the
 % target key before we can conclude that functions on dicts
@@ -667,6 +710,7 @@ safe_primitive('$dicts':'.'(_,K,_)) :-
     ).
 
 dict_built_in(get(_)).
+dict_built_in(get(_,_)).
 dict_built_in(put(_)).
 dict_built_in(put(_,_)).
 
@@ -680,6 +724,7 @@ safe_primitive(system:sort(_,_,_,_)).
 safe_primitive(system:between(_,_,_)).
 safe_primitive(system:succ(_,_)).
 safe_primitive(system:plus(_,_,_)).
+safe_primitive(system:float_class(_,_)).
 safe_primitive(system:term_variables(_,_)).
 safe_primitive(system:term_variables(_,_,_)).
 safe_primitive(system:'$term_size'(_,_,_)).
@@ -692,6 +737,7 @@ safe_primitive(system:upcase_atom(_,_)).
 safe_primitive(system:is_list(_)).
 safe_primitive(system:memberchk(_,_)).
 safe_primitive(system:'$skip_list'(_,_,_)).
+safe_primitive(system:'$seek_list'(_, _, _, _)).
                                         % attributes
 safe_primitive(system:get_attr(_,_,_)).
 safe_primitive(system:get_attrs(_,_)).
@@ -729,6 +775,12 @@ stack_name(local).
 stack_name(trail).
 
 safe_primitive('$tabling':abolish_all_tables).
+safe_primitive('$tabling':'$wrap_tabled'(Module:_Head, _Mode)) :-
+    prolog_load_context(module, Module),
+    !.
+safe_primitive('$tabling':'$moded_wrap_tabled'(Module:_Head,_,_,_,_)) :-
+    prolog_load_context(module, Module),
+    !.
 
 
 % use_module/1.  We only allow for .pl files that are loaded from
@@ -736,6 +788,9 @@ safe_primitive('$tabling':abolish_all_tables).
 
 safe_primitive(system:use_module(Spec, _Import)) :-
     safe_primitive(system:use_module(Spec)).
+safe_primitive(system:load_files(Spec, Options)) :-
+    safe_primitive(system:use_module(Spec)),
+    maplist(safe_load_file_option, Options).
 safe_primitive(system:use_module(Spec)) :-
     ground(Spec),
     (   atom(Spec)
@@ -767,6 +822,12 @@ segments_to_path(X) -->
 
 save_extension(pl).
 
+safe_load_file_option(if(changed)).
+safe_load_file_option(if(not_loaded)).
+safe_load_file_option(must_be_module(_)).
+safe_load_file_option(optimise(_)).
+safe_load_file_option(silent(_)).
+
 %!  safe_assert(+Term) is semidet.
 %
 %   True if assert(Term) is safe,  which   means  it  asserts in the
@@ -807,6 +868,15 @@ safe_global_var(Name) :-
 %
 %   Declare the given global variable safe to write to.
 
+%!  safe_current_predicate(+X)
+%
+%   current_predicate/1 is safe when not used with qualification.
+
+safe_current_predicate(X) :-
+    nonvar(X),
+    X = _:_, !,
+    fail.
+safe_current_predicate(_).
 
 %!  safe_meta(+Goal, -Called:list(callable)) is semidet.
 %
@@ -832,6 +902,9 @@ safe_meta(system:format(Output, Format, Args), Calls) :-
     format_calls(Format, Args, Calls).
 safe_meta(prolog_debug:debug(_Term, Format, Args), Calls) :-
     format_calls(Format, Args, Calls).
+safe_meta(system:set_prolog_flag(Flag, Value), []) :-
+    atom(Flag),
+    safe_prolog_flag(Flag, Value).
 safe_meta('$attvar':freeze(_Var,Goal), [Goal]).
 safe_meta(phrase(NT,Xs0,Xs), [Goal]) :- % phrase/2,3 and call_dcg/2,3
     expand_nt(NT,Xs0,Xs,Goal).
@@ -845,6 +918,8 @@ safe_meta('$tabling':abolish_table_subgoals(V), []) :-
     \+ qualified(V).
 safe_meta('$tabling':current_table(V, _), []) :-
     \+ qualified(V).
+safe_meta('$tabling':tnot(G), [G]).
+safe_meta('$tabling':not_exists(G), [G]).
 
 qualified(V) :-
     nonvar(V),
@@ -1002,6 +1077,7 @@ safe_meta(system:setup_call_catcher_cleanup(0,0,*,0)).
 safe_meta('$attvar':call_residue_vars(0,*)).
 safe_meta('$syspreds':call_with_inference_limit(0,*,*)).
 safe_meta('$syspreds':call_with_depth_limit(0,*,*)).
+safe_meta('$syspreds':undo(0)).
 safe_meta(^(*,0)).
 safe_meta(\+(0)).
 safe_meta(call(0)).
@@ -1013,6 +1089,7 @@ safe_meta(call(5,*,*,*,*,*)).
 safe_meta(call(6,*,*,*,*,*,*)).
 safe_meta('$tabling':start_tabling(*,0)).
 safe_meta('$tabling':start_tabling(*,0,*,*)).
+safe_meta(wfs:call_delays(0,*)).
 
 %!  safe_output(+Output)
 %
@@ -1038,16 +1115,16 @@ safe_output(current_error).
 
 :- public format_calls/3.                       % used in pengines_io
 
-format_calls(Format, _Args, _Calls) :-
-    var(Format),
-    !,
-    instantiation_error(Format).
 format_calls(Format, Args, Calls) :-
+    is_list(Args),
+    !,
     format_types(Format, Types),
     (   format_callables(Types, Args, Calls)
     ->  true
     ;   throw(error(format_error(Format, Types, Args), _))
     ).
+format_calls(Format, Arg, Calls) :-
+    format_calls(Format, [Arg], Calls).
 
 format_callables([], [], []).
 format_callables([callable|TT], [G|TA], [G|TG]) :-
@@ -1124,6 +1201,7 @@ safe_pattr(multifile(_)).
 safe_pattr(public(_)).
 safe_pattr(meta_predicate(_)).
 safe_pattr(table(_)).
+safe_pattr(non_terminal(_)).
 
 safe_pattr(Var, _) :-
     var(Var),
@@ -1153,13 +1231,14 @@ safe_source_directive(op(_,_,Name)) :-
 safe_source_directive(set_prolog_flag(Flag, Value)) :-
     !,
     atom(Flag), ground(Value),
-    safe_directive_flag(Flag, Value).
+    safe_prolog_flag(Flag, Value).
 safe_source_directive(style_check(_)).
 safe_source_directive(initialization(_)).   % Checked at runtime
 safe_source_directive(initialization(_,_)). % Checked at runtime
 
 directive_loads_file(use_module(library(X)), X).
 directive_loads_file(use_module(library(X), _Imports), X).
+directive_loads_file(load_files(library(X), _Options), X).
 directive_loads_file(ensure_loaded(library(X)), X).
 directive_loads_file(include(X), X).
 
@@ -1183,19 +1262,43 @@ safe_path(A/B) :-
     safe_path(B).
 
 
-%!  safe_directive_flag(+Flag, +Value) is det.
+%!  safe_prolog_flag(+Flag, +Value) is det.
 %
-%   True if it is safe to set the flag Flag in a directive to Value.
+%   True if it is safe to set the flag Flag to Value.
 %
 %   @tbd    If we can avoid that files are loaded after changing
 %           this flag, we can allow for more flags.  The syntax
 %           flags are safe because they are registered with the
 %           module.
 
-safe_directive_flag(generate_debug_info, _).
-safe_directive_flag(var_prefix, _).
-safe_directive_flag(double_quotes, _).
-safe_directive_flag(back_quotes, _).
+% misc
+safe_prolog_flag(generate_debug_info, _).
+safe_prolog_flag(optimise, _).
+safe_prolog_flag(occurs_check, _).
+safe_prolog_flag(write_attributes, _).
+% syntax
+safe_prolog_flag(var_prefix, _).
+safe_prolog_flag(double_quotes, _).
+safe_prolog_flag(back_quotes, _).
+safe_prolog_flag(rational_syntax, _).
+% arithmetic
+safe_prolog_flag(prefer_rationals, _).
+safe_prolog_flag(float_overflow, _).
+safe_prolog_flag(float_zero_div, _).
+safe_prolog_flag(float_undefined, _).
+safe_prolog_flag(float_underflow, _).
+safe_prolog_flag(float_rounding, _).
+safe_prolog_flag(float_rounding, _).
+safe_prolog_flag(max_rational_size, _).
+safe_prolog_flag(max_rational_size_action, _).
+% tabling
+safe_prolog_flag(max_answers_for_subgoal,_).
+safe_prolog_flag(max_answers_for_subgoal_action,_).
+safe_prolog_flag(max_table_answer_size,_).
+safe_prolog_flag(max_table_answer_size_action,_).
+safe_prolog_flag(max_table_subgoal_size,_).
+safe_prolog_flag(max_table_subgoal_size_action,_).
+
 
 %!  prolog:sandbox_allowed_expansion(:G) is det.
 %
@@ -1210,13 +1313,10 @@ safe_directive_flag(back_quotes, _).
 %   and we only need to be  careful   if  the sandboxed code defines
 %   expansion rules.
 
-prolog:sandbox_allowed_expansion(Directive) :-
-    prolog_load_context(module, M),
-    debug(sandbox(expansion), 'Expand in ~p: ~p', [M, Directive]),
-    fail.
 prolog:sandbox_allowed_expansion(M:G) :-
     prolog_load_context(module, M),
     !,
+    debug(sandbox(expansion), 'Expand in ~p: ~p', [M, G]),
     safe_goal(M:G).
 prolog:sandbox_allowed_expansion(_,_).
 
